/* XPENGUINS - cool little penguins that walk along the tops of your windows
 * Copyright (C) 1999, 2000  Robin Hogan
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * XPENGUINS is a kind of a cross between lemmings and xsnow - penguins
 * fall from the top of the screen and walk across the tops of windows.
 * The images are take from `Pingus', a lemmings-type game for Linux.
 * All the nasty X calls are in `toon.c', so xpenguins.c should be easily
 * modifiable to animate any small furry (or feathery) animal of your choice.
 * 
 * Robin Hogan <R.J.Hogan@reading.ac.uk>
 * Project started 22 November 1999
 * Last update 26 February 2000
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

/* C structures defined here */
#include "toon.h"

/*
 * The penguin shapes are defined here - includes links to all the xpm images.
 * If you want to have something else falling from the top of your screen (the
 * possibilities are endless!) then point the following line somewhere else. 
 */
#include "penguins/def.h"

#define MAX_PENGUINS 256
#define DEFAULT_DELAY 50
#define RandInt(maxint) ((int) ((maxint)*((float) rand()/(RAND_MAX+1.0))))

#define XPENGUINS_VERSION "1.1"
#define XPENGUINS_AUTHOR "Robin Hogan"
#define XPENGUINS_DATE "26 February 2000"

int finished=0;
int new_positions=0;
int verbose=1;

void InitPenguin(Toon *penguin) {
   SetToonType(penguin, PENGUIN_FALLER,PENGUIN_FOREWARD);
   SetToonPosition(penguin, RandInt(ToonDisplayWidth()-32), 0);
   SetToonVelocity(penguin, RandInt(2)*2-1, 3);
}

void ShowUsage(); 
void ShowUsage() {
   fprintf(stdout,"Valid options are:\n");
   fprintf(stdout,"   -display screen:0.0      Send the penguins to the display `screen:0.0'\n");
   fprintf(stdout,"   -delay millisecs         Set the delay between frames (default %d)\n",
         DEFAULT_DELAY);
   fprintf(stdout,"   -penguins n              Create n penguins (max %d)\n",
         MAX_PENGUINS);
   fprintf(stdout,"   -ignorepopups            Penguins fall through `popup' windows\n");
   fprintf(stdout,"   -q -quiet --quiet        Don't announce explosion of penguins on exit\n");
   fprintf(stdout,"   -v -version --version    Show version information\n");
   fprintf(stdout,"   -h -? -help --help       Show this message\n");
}

int main (int argc, char **argv) {
   char *display_name=NULL;
   Toon penguin[MAX_PENGUINS];
   int prefd[MAX_PENGUINS]; /* preferred direction; -1 means none */
   char input[100], ignore_popups=0;
   int windows_moved,status,npenguins=8,i,n;
   unsigned long sleep_usec=DEFAULT_DELAY*1000;

/* Handle command-line arguments */
   for (n=1;n<argc;n++) {
      if (strcmp(argv[n],"-penguins") == 0) {
         if (argc > ++n) {
            npenguins=atoi(argv[n]);
            if (npenguins > MAX_PENGUINS) {
               fprintf(stderr,"Warning: only %d penguins created\n",
                     npenguins=MAX_PENGUINS);
            }
            else if (npenguins <= 0) {
               fprintf(stderr,"Warning: no penguins created\n",
                     npenguins=0);
            }
         }
         else {
            fprintf(stderr,"Error: number of penguins not specified\n");
            ShowUsage();
            exit(1);
         }
      }
      else if (strcmp(argv[n],"-delay") == 0) {
         if (argc > ++n) {
            sleep_usec=1000*atoi(argv[n]);
         }
         else {
            fprintf(stderr,"Error: delay in milliseconds not specified\n");
            ShowUsage();
            exit(1);
         }
      }
      else if (strcmp(argv[n],"-display") == 0) {
         if (argc > ++n) {
            display_name=argv[n];
         }
         else {
            fprintf(stderr,"Error: display not specified\n");
            ShowUsage();
            exit(1);
         }
      }
      else if (strcmp(argv[n],"-ignorepopups") == 0 ) {
         ignore_popups=1;
      }
      else if (strcmp(argv[n],"-h") == 0 || strcmp(argv[n],"-help") == 0 || 
               strcmp(argv[n],"-?") == 0 || strcmp(argv[n],"--help") == 0) {
         fprintf(stdout,"XPenguins version %s (%s) by %s\n",
               XPENGUINS_VERSION, XPENGUINS_DATE, XPENGUINS_AUTHOR);
         ShowUsage();
         exit(0);
      }
      else if (strcmp(argv[n],"-v") == 0 || strcmp(argv[n],"-version") == 0 || 
               strcmp(argv[n],"--version") == 0) {
         fprintf(stdout,"XPenguins version %s\n", XPENGUINS_VERSION);
         exit(0);
      }
      else if (strcmp(argv[n],"-q") == 0 || strcmp(argv[n],"-quiet") == 0 || 
               strcmp(argv[n],"--quiet") == 0) {
         verbose=0;
      }
      else {
         fprintf(stderr,"Warning: `%s' not understood - use `-h' for a list of options\n",argv[n]);
      }
   }

/* reset random-number generator */
   srand(time((long *) NULL));

/* contact X server and set up some basic X stuff */
   if (OpenToonDisplay(display_name) == NULL) {
      fprintf(stderr,"Error: %s\n", ToonErrorMessage());
      exit(1);
   };
/* Set up various preferences: Edge of screen is solid, and if a signal is caught
 * then exit the main event loop */
   if (ignore_popups)
      ToonConfigure(TOON_EDGEBLOCK | TOON_CATCHSIGNALS | TOON_NOSOLIDPOPUPS);
   else
      ToonConfigure(TOON_EDGEBLOCK | TOON_CATCHSIGNALS | TOON_SOLIDPOPUPS);
/* Send the pixmaps to the X server  - penguin_data should have been 
 * defined in penguins/def.h */
   InstallToonData(penguin_data,PENGUIN_TYPES);

/* initialise penguins */
   for (i=0;i<npenguins;i++) {
      InitPenguin(penguin+i);
      prefd[i]=-1;
   }

   DrawToons(penguin,npenguins);
   FlushToons();

/* Find out where the windows are - should be done just before beginning the 
 * event loop */
   ToonLocateWindows();
/* Event loop */
   while (!finished) {
      /* check if windows have moved, and flush the display */
      windows_moved=ToonWindowsMoved();
      /* if so, check for squashed toons */
      if (windows_moved) {
         for (i=0;i<npenguins;i++) {
            if (penguin[i].active && ToonBlocked(penguin+i,TOON_HERE)) {
               SetToonType(penguin+i,PENGUIN_EXPLOSION,
                     PENGUIN_FOREWARD);
               SetToonPosition(penguin+i, penguin[i].x-8, penguin[i].y-8);
            }
         }
      }
      /* pause */
      ToonSleep(sleep_usec);
      for (i=0;i<npenguins;i++) {
         if (!(penguin[i].active)) {
            InitPenguin(penguin+i);
            break;
         }
         else {
            status=AdvanceToon(penguin+i,TOON_MOVE);
         }
         switch (penguin[i].type) {
            case PENGUIN_FALLER:
               if (status != TOON_OK) {
                  if (ToonBlocked(penguin+i,TOON_DOWN)) {
                     int direction;
                     if (prefd[i]>-1)
                        direction=prefd[i];
                     else
                        direction=RandInt(2);
                     SetToonType(penguin+i, PENGUIN_WALKER,direction);
                     SetToonVelocity(penguin+i, 4*((2*direction)-1), 0);
                     prefd[i]=-1;
                  }
                  else {
                     if (RandInt(2)) {
                        SetToonVelocity(penguin+i,-penguin[i].u,3);
                     }
                        SetToonType(penguin+i, PENGUIN_CLIMBER, penguin[i].u>0);
                        SetToonVelocity(penguin+i,0,-4);
                  }
               }
               break;
            case PENGUIN_TUMBLER:
               if (status != TOON_OK) {
                  int direction;
                  if (prefd[i]>-1)
                     direction=prefd[i];
                  else
                     direction=RandInt(2);
                  SetToonType(penguin+i, PENGUIN_WALKER,direction);
                  SetToonVelocity(penguin+i, 4*((2*direction)-1), 0);
                  prefd[i]=-1;
               }
               break;
            case PENGUIN_WALKER:
               if (!ToonBlocked(penguin+i,TOON_DOWN)) {
                  prefd[i]=penguin[i].direction;
                  SetToonType(penguin+i, PENGUIN_TUMBLER,
                        PENGUIN_FOREWARD);
                  SetToonVelocity(penguin+i, 0, 6);
               }
               else if (status != TOON_OK) {
                  /* Blocked! We can turn round, fly or climb... */
                  switch (RandInt(8)) {
                     case 0:
                        SetToonType(penguin+i,PENGUIN_FLOATER,
                              PENGUIN_FOREWARD);
                        SetToonVelocity(penguin+i,RandInt(5)*(-penguin[i].u/4),-3);
                        break;
                     case 1:
                        SetToonType(penguin+i, PENGUIN_CLIMBER,
                              penguin[i].direction);
                        SetToonVelocity(penguin+i,0,-4);
                        break;
                     default:
                        SetToonType(penguin+i, PENGUIN_WALKER,
                              !penguin[i].direction);
                        SetToonVelocity(penguin+i,-penguin[i].u,0);
                  }
               }
               break;
            case PENGUIN_CLIMBER:
               if (!ToonBlocked(penguin+i,penguin[i].direction)) {
                  SetToonType(penguin+i, PENGUIN_WALKER,penguin[i].direction);
                  SetToonVelocity(penguin+i, 4*((2*penguin[i].direction)-1), 0);
                  SetToonPosition(penguin+i, penguin[i].x+(2*penguin[i].direction)-1,
                        penguin[i].y);
                  prefd[i]=penguin[i].direction;
               }
               else if (status != TOON_OK) {
                  SetToonVelocity(penguin+i, (1-penguin[i].direction*2), 3);
                  SetToonType(penguin+i, PENGUIN_FALLER,PENGUIN_FOREWARD);
               }
               break;
            case PENGUIN_FLOATER:
               if (status != TOON_OK) {
                  if (ToonBlocked(penguin+i,TOON_UP)) {
                     SetToonVelocity(penguin+i, ((penguin[i].u>0)*2-1), 3);
                     SetToonType(penguin+i, PENGUIN_FALLER,
                           PENGUIN_FOREWARD);
                  }
                  else {
                     SetToonVelocity(penguin+i,-penguin[i].u, -3);
                  }
               }
               break;
            case PENGUIN_EXPLOSION:
               penguin[i].active=0;
         }
      }
      /* First erase them all, then draw them all - should reduce flickering */
      EraseToons(penguin,npenguins);
      DrawToons(penguin,npenguins);
      /* Has an interrupt signal been received? If so, quit gracefully */
      finished=ToonSignal();
   }

   /* Exit sequence... */
   /* Any more signals (TERM, HUP or INT) and the penguins are erased immediately */
   ToonConfigure(TOON_EXITGRACEFULLY);
   if (1) {
      /* Nice exit sequence... */
      if (verbose) fprintf(stderr,"Interrupt received: exploding penguins");
      for (i=0;i<npenguins;i++) {
         SetToonType(penguin+i,PENGUIN_BOMBER,PENGUIN_FOREWARD);
      }
      for (n=0;n<penguin_data[PENGUIN_BOMBER].nframes;n++) {
         EraseToons(penguin,npenguins);
         DrawToons(penguin,npenguins);
         FlushToons();
         for (i=0;i<npenguins;i++) {
            AdvanceToon(penguin+i,TOON_FORCE);
         }
         ToonSleep(sleep_usec);
         if (verbose) fprintf(stderr,".");
      }
      if (verbose) fprintf(stderr,"done\n");
   }
   EraseToons(penguin,npenguins);
   CloseToonDisplay();
   exit(0);
}

